/*
 * Copyright (C) 2007 Richard Hughes <richard@hughsie.com>
 *
 * Licensed under the GNU Lesser General Public License Version 2
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include <gmodule.h>
#include <glib.h>
#include <glib/gstdio.h>
#include <glib/gprintf.h>
#include <ohm-plugin.h>
#include <time.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <string.h>

enum {
	CONF_AC_STATE_CHANGED,
	CONF_BACKLIGHT_STATE_CHANGED,
	CONF_LID_STATE_CHANGED,
	CONF_TABLET_STATE_CHANGED,
	CONF_BRIGHTNESS_AC_CHANGED,
	CONF_BRIGHTNESS_BATTERY_CHANGED,
	CONF_BRIGHTNESS_IDLE_CHANGED,
	CONF_TIMEOUTS_POWERSAVE_CHANGED,
	CONF_TIMEOUTS_MOMENTARY_CHANGED,
	CONF_LAST
};

static gint
display_get_brightness (OhmPlugin *plugin)
{
	gint level = -1;
	FILE *fp;
	char levelchar[4];
	
	g_debug("%s", G_STRFUNC);
	fp = g_fopen("/sys/class/backlight/dcon-bl/brightness", "r");
	if (!fp) {
		g_warning ("Couldn't open /sys/class/backlight/dcon-bl/brightness.");
		return -1;
	}

	while (fgets(levelchar, sizeof(levelchar), fp) != NULL) {
		g_debug ("Level is %s.", levelchar);
		level = atoi(levelchar);
		g_debug ("Level is %d.", level);
		break;
	}

	fclose(fp);
	
	return level;
}

static void
reset_brightness (OhmPlugin *plugin)
{
	gint onac;
	gint value_current;
	gint value_ac;
	gint value_battery;
	gint value_idle;
	gint value_new = -1;
	gint value_user;
	
	/* FIXME: turn on dcon -- why was this here? */
	/* ohm_plugin_conf_set_key (plugin, "backlight.state", 1); */

	g_debug("%s", G_STRFUNC);

	ohm_plugin_conf_get_key (plugin, "acadapter.state", &onac);
	ohm_plugin_conf_get_key (plugin, "display.value_ac", &value_ac);
	ohm_plugin_conf_get_key (plugin, "display.value_battery", &value_battery);
	ohm_plugin_conf_get_key (plugin, "display.value_idle", &value_idle);
	ohm_plugin_conf_get_key (plugin, "display.value_user", &value_user);

	value_current = display_get_brightness(plugin);

	if (value_current == -1) {
		g_warning("Can't read brightness!");
		return;
	}

	if ((value_current != value_ac) && (value_current != value_battery)
	   && (value_current != value_idle)) {
		g_debug ("Setting value_new to %d, ac: %d, battery: %d, idle: %d.", value_current, value_ac, value_battery, value_idle);
		value_new = value_current;
		ohm_plugin_conf_set_key (plugin, "display.value_user", value_new);
	}
	else {
		/* Do we need to restore from user-set brightness? */
		ohm_plugin_conf_get_key (plugin, "display.value_user", &value_user);
		if ((value_user != value_ac) && (value_user != value_battery)
		   && (value_user != value_idle)) {
			value_new = value_user;
			g_debug ("value_user (%d) isn't a default, so preserving.", value_user);
		}
		else {
			if (onac == TRUE) {
				value_new = value_ac;
			} else {
				value_new = value_battery;
			}
		}
	}
	
	g_debug("%s: acadapter.state = %d, setting brightness %d", G_STRFUNC, onac, value_new);

	/* dim screen to idle brightness */
	ohm_plugin_conf_set_key (plugin, "backlight.hardware_brightness", value_new);
}

static gboolean
cpu_is_busy (OhmPlugin *plugin)
{
#define IDLE_PERCENT 20

	gchar *output = NULL;
	GError *error = NULL;
	gint idle;

	g_debug("in %s", G_STRFUNC);

	if (!g_spawn_command_line_sync ("/bin/sh -c \"vmstat 1 2 | tail -n1 | awk '{print $15}'\"", &output, NULL, NULL, &error)) {
		g_warning ("Error spawning: %s\n", error->message);
		g_error_free (error);
	}

	idle = atoi(output);
	if (idle < IDLE_PERCENT) {
		g_debug("idle %d is less than IDLE_PERCENT; inhibiting", idle);
		return TRUE;
	} else {
		g_debug("idle %d is more than IDLE_PERCENT; proceeding", idle);
	}

	return FALSE;
}

static void 
reset_idle(OhmPlugin *plugin)
{
	gint new_timeout;
	g_debug("in %s", G_STRFUNC);
	
	/* We've resumed now.  Bring the backlight back immediately. */
	ohm_plugin_conf_set_key (plugin, "backlight.state", 1);
	ohm_plugin_conf_set_key (plugin, "idle.reset", 1);

	ohm_plugin_conf_get_key (plugin, "timeouts.timer_momentary", &new_timeout);
	ohm_plugin_conf_set_key (plugin, "idle.timeout", new_timeout);
	ohm_plugin_conf_set_key (plugin, "timeouts.momentary", 0);
	ohm_plugin_conf_set_key (plugin, "timeouts.powerdown", 0);
}

/* todo, we need to inhibit the screen from dimming */
static void
brightness_momentary (OhmPlugin *plugin, gboolean is_idle)
{
#define TTY1 "/dev/tty1"
#define TTY2 "/dev/tty2"

	gint lidshut;
	gint value_ac;
	gint value_idle;
	gint value_current;
	gint state;
	gint reset;
	struct stat tty1;
	struct stat tty2;
	time_t curtime;
	
	g_debug("%s", G_STRFUNC);
	reset = 0;

	ohm_plugin_conf_get_key (plugin, "backlight.state", &state);
	if (state == 0) {
                g_debug("%s: state is 0", G_STRFUNC);
		/* work round a idletime bugs */
		return;
	}

	/* if lid shut */
	ohm_plugin_conf_get_key (plugin, "button.lid", &lidshut);
	if (lidshut == 1) {
	        g_debug("%s: lid shut", G_STRFUNC);
		/* we've already turned off the screen */
		return;
	}

	/* if not idle any more */
	if (is_idle == FALSE) {
		g_debug("%s: is_idle=FALSE< restting brighness", G_STRFUNC);
		reset_brightness (plugin);
		return;
	}

	curtime = time(NULL);
	g_debug ("time now is %lu", (unsigned long) curtime);

	if (stat(TTY1, &tty1) >= 0) {
		g_debug ("tty1 time is %lu", (unsigned long) tty1.st_mtime);
		if (curtime-tty1.st_mtime < 60) {
			g_debug("tty1 is in use; inhibiting screen dim.");
			reset = 1;
		}
	}

	if (stat(TTY2, &tty2) >= 0) {
		g_debug ("tty2 time is %lu", (unsigned long) tty2.st_mtime);
		if (curtime-tty2.st_mtime < 60) {
			g_debug("tty2 is in use; inhibiting screen dim.");
			reset = 1;
		}
	}

	if (cpu_is_busy(plugin)) {
		g_debug("cpu busy; inhibiting screen dim.");
		reset = 1;
	}

	if (reset == 1) {
		reset_idle(plugin);
		return;
	}

	/* dim screen to idle brightness */

	/* First, see whether the current brightness is user-set. */ 
	value_current = display_get_brightness(plugin);
	
	if (value_current == -1) {
		g_warning ("Can't read brightness!");
		return;
	}
	
	ohm_plugin_conf_get_key (plugin, "display.value_ac", &value_ac);
	ohm_plugin_conf_get_key (plugin, "display.value_idle", &value_idle);

	if ((value_current != value_ac) && (value_current != value_idle)) {
		g_debug ("Brightness is %d and AC is %d and idle is %d.", value_current, value_ac, value_idle);
		ohm_plugin_conf_set_key (plugin, "display.value_user", value_current);
	}

	if (value_current > value_idle) {  
		g_debug ("User brightness %d is > idle brightness %d", value_current, value_idle);
		g_debug ("%s: Setting brightness to %d", G_STRFUNC, value_idle);
		ohm_plugin_conf_set_key (plugin, "backlight.hardware_brightness", value_idle);
	}
	else {
		g_debug ("User brightness %d is < idle brightness %d", value_current, value_idle);
		g_debug ("Not dimming because value_user %d is already less than value_idle %d.", value_current, value_idle);
	}
}

/* todo, we need to inhibit the screen from turning off */
static void
backlight_powersave (OhmPlugin *plugin, gboolean is_idle)
{
	gint lidshut;
	gint state;

	ohm_plugin_conf_get_key (plugin, "backlight.state", &state);
	if (is_idle && state == 0) {
	        g_debug("%s: is_idle 0", G_STRFUNC);
		/* work round a idletime bugs */
		return;
	}

	/* if lid shut */
	ohm_plugin_conf_get_key (plugin, "button.lid", &lidshut);
	if (lidshut == 1) {
	        g_debug("%s: lidshut 1", G_STRFUNC);
		/* we've already turned off the screen */
		return;
	}

	/* if not idle any more */
	if (is_idle == FALSE) {
	        g_debug("%s: if not idle any more", G_STRFUNC);
		ohm_plugin_conf_set_key (plugin, "backlight.state", 1);
		reset_brightness (plugin);
		return;
	}

}

static void
lid_closed (OhmPlugin *plugin, gboolean is_closed)
{
	/* just turn on DCON */
	if (is_closed == FALSE) {
		ohm_plugin_conf_set_key (plugin, "backlight.state", 1);
		return;
	}
		
	/* just turn off dcon unconditionally */
	ohm_plugin_conf_set_key (plugin, "backlight.state", 0);
}

/**
 * plugin_notify:
 * @plugin: This class instance
 *
 * Notify the plugin that a key marked with ohm_plugin_conf_interested ()
 * has it's value changed.
 * An enumerated numeric id rather than the key is returned for processing speed.
 */
static void
plugin_notify (OhmPlugin *plugin, gint id, gint value)
{
	switch (id) {
	case CONF_BRIGHTNESS_AC_CHANGED:
	case CONF_BRIGHTNESS_BATTERY_CHANGED:
	case CONF_BRIGHTNESS_IDLE_CHANGED:
		reset_brightness (plugin);
		break;
	case CONF_TIMEOUTS_POWERSAVE_CHANGED:
		backlight_powersave (plugin, (value == 1));
		break;
	case CONF_LID_STATE_CHANGED:
		lid_closed (plugin, (value == 1));
		break;
	case CONF_TABLET_STATE_CHANGED:
		break;
	case CONF_AC_STATE_CHANGED:
	case CONF_BACKLIGHT_STATE_CHANGED:
		reset_brightness (plugin);
		break;
	case CONF_TIMEOUTS_MOMENTARY_CHANGED:
		brightness_momentary (plugin, (value == 1));
		break;
	}
}

OHM_PLUGIN_DESCRIPTION (
	"Display",			/* description */
	"0.0.1",			/* version */
	"richard@hughsie.com",		/* author */
	OHM_LICENSE_LGPL,		/* license */
	NULL,				/* initialize */
	NULL,				/* destroy */
	plugin_notify			/* notify */
);

OHM_PLUGIN_SUGGESTS (
	"timeouts",
	"acadapter",
	"buttons",
	"xrandr",
	"backlight"
);

OHM_PLUGIN_INTERESTED (
	{"acadapter.state", CONF_AC_STATE_CHANGED},
	{"backlight.state", CONF_BACKLIGHT_STATE_CHANGED},
	{"button.lid", CONF_LID_STATE_CHANGED},
	{"button.tablet", CONF_TABLET_STATE_CHANGED},
	{"timeouts.powersave", CONF_TIMEOUTS_POWERSAVE_CHANGED},
	{"timeouts.momentary", CONF_TIMEOUTS_MOMENTARY_CHANGED},
	{"display.value_ac", CONF_BRIGHTNESS_AC_CHANGED},
	{"display.value_battery", CONF_BRIGHTNESS_BATTERY_CHANGED},
	{"display.value_idle", CONF_BRIGHTNESS_IDLE_CHANGED}
);
